class PathList(object):

    def __init__(
            self,
            pathstring,
            molecule_scenario_directory):
        self._molecule_scenario_directory = molecule_scenario_directory
        self._pathlist = self._pathstring_to_pathlist_(pathstring)

    def get(self):
        return self._pathlist

    def _get_molecule_scenario_directory_(self):
        return self._molecule_scenario_directory

    def _pathstring_to_pathlist_(self, pathstring):
        '''Returns list of yaml files with extra vars'''
        pathlist = list()
        base_dir = self._get_molecule_scenario_directory_()
        for item in str(pathstring).split(':'):
            path = base_dir / item
            path = path.resolve()
            if path.is_file():
                pathlist.append(path)
            if path.is_dir():
                filelist = path.glob('**/*.yml')
                for file in filelist:
                    pathlist.append(file.resolve())

        return sorted(pathlist)



import json


class MultiTestVars(object):
    '''Expose ansible variabless of a molecule scenario.

    Include ansible facts form molecule host.
    Include roles from ansible project directory.
    Resolve jinja2 template variables.
    '''

    def __init__(
            self,
            ansibleinventory,
            moleculelog,
            debug_jsonvars,
            moleculebook,
            jsonvars,
            gather_facts,
            extra_vars):

        # this variable will be returned by the multitestvars fixture
        self._multitestvars = dict()

        for ansiblehost in ansibleinventory.hosts:
            host = str(ansiblehost)

            # get ansible variables
            testvars_unresolved = moleculebook.get_vars(
                gather_facts,
                extra_vars,
                host)

            # convert python variables to json
            testvars_unresolved_json = json.dumps(testvars_unresolved)

            # set jsonvars to unresolved testvars
            jsonvars.set(testvars_unresolved_json)

            # resolve unresolved json testvars
            jsonvars.resolve()

            # print jsonvars debug info if command line flag is specified
            if debug_jsonvars:
                moleculelog.debug(jsonvars.get_debug())

            # get resolved testvars from jsonvars
            testvars_resolved_json = jsonvars.get()

            # convert json vars to python
            testvars = json.loads(testvars_resolved_json)

            # save testvars
            self._multitestvars[host] = testvars

    def get_multitestvars(self):
        return self._multitestvars

    def get_cache(request, cache_key):
        try:
            # read testvars from cache
            # you can enable cache support in molecule.yml:
            # molecule -> verifier -> options
            # option "p: cacheprovider"
            multitestvars = request.config.cache.get(cache_key, None)
        except AttributeError:
            multitestvars = None
        return multitestvars

    def set_cache(self, request, cache_key):
        try:
            request.config.cache.set(cache_key, self._testvars)
        except AttributeError:
            pass



import json
import re


class JsonVars(object):

    def __init__(self,
                 templates,
                 gather_molecule):

        # jinja2 templates
        self._templates = templates

        # unresolved json variables
        self._jsonvars = dict()

        # should we ignore templates containing MOLECULE_ or molecule_file?
        if gather_molecule:
            r = r'(["])?{{(.*?)}}(["])?'
        else:
            r = r'(["])?{{((?:(?!.(?:MOLECULE_|molecule_file)).)*?)}}(["])?'

        # compile regular expression to find templates
        self._regex_templates = re.compile(r)

        # trivial hash table with templates as hash values
        self._hash_table = list()

        # cache table
        self._templates_lookup_table = list()

        self._spots = list()

    def get(self):
        return self._jsonvars

    def get_debug(self):
        msg = '\n'
        msg += self._debug_hash_table_()
        msg += '\n'
        msg += self._debug_templates_lookup_table_()
        msg += '\n'
        msg += self._debug_templates_()
        msg += '\n'
        msg += self._debug_spots_()
        msg += '\n'
        msg += self._debug_jsonvars_()
        msg += '\n'
        return msg

    def reset(self):
        self._templates.reset()
        self._jsonvars = dict()
        self._hash_table = list()
        self._templates_lookup_table = list()
        self._spots = list()

    def resolve(self):

        # first part of query / replace
        self._query_templates_()

        # run a large playbook against the molecule host
        # to resolve all jinja2 templates in one run
        self._templates.resolve()

        # second part of query / replace
        self._replace_templates_()

    def set(self, jsonvars):
        self._jsonvars = jsonvars

    def _query_templates_(self):
        '''Return all unresolved jinja2 templates.'''

        # find all templates in json variables string
        templates_unresolved = \
            self._regex_templates.findall(self._jsonvars)

        # create hash table so that we don't resolve templates twice
        for template_unresolved in templates_unresolved:
            self._hash_table.append(template_unresolved[1])

        for index, template_unresolved in enumerate(templates_unresolved):
            spot = dict()

            # save template spot environment
            if template_unresolved[0]:
                spot['left_quote'] = True
            else:
                spot['left_quote'] = False
            if template_unresolved[2]:
                spot['right_quote'] = True
            else:
                spot['right_quote'] = False
            self._spots.append(spot)

            # get first occurence of our template
            first = self._hash_table.index(template_unresolved[1])

            # check if this is a double template
            if first < index:

                # existing template
                reference = self._templates_lookup_table[first]
                self._templates_lookup_table.append(reference)

            else:

                # new template
                self._templates_lookup_table.append(
                    len(self._templates.get_templates()))
                self._templates.add(template_unresolved[1].strip())

    def _replace_templates_(self):
        '''Replace jinja2 templates by resolved templates.'''

        # keep track of the position in self._templates_resolved
        self._resolve_var_index_ = 0

        self._jsonvars = \
            self._regex_templates.sub(lambda x: self._resolve_template_(),
                                      self._jsonvars)

    def _resolve_template_(self):
        '''Replace jinja2 template by resolved template.'''
        spot = self._spots[self._resolve_var_index_]
        index = self._templates_lookup_table[self._resolve_var_index_]
        template = self._templates.get_template(index)

        template_resolved = template['resolved'].strip('"')
        if (template['string'] and spot['left_quote']) \
                or (spot['left_quote'] and not spot['right_quote']):
            template_resolved = '"' + template_resolved
        if (template['string'] and spot['right_quote']) \
                or (spot['right_quote'] and not spot['left_quote']):
            template_resolved = template_resolved + '"'
        self._resolve_var_index_ += 1

        return template_resolved

    def _debug_hash_table_(self):
        msg = '+++ hash_table +++\n'
        for index, hash in enumerate(self._hash_table):
            msg += 'hash ' + str(index) + ' -> ' + str(hash) + '\n'
        return msg

    def _debug_templates_lookup_table_(self):
        msg = '+++ lookup_table +++\n'
        for index, lookup in enumerate(self._templates_lookup_table):
            msg += str(index) + ' -> ' + str(lookup) + '\n'
        return msg

    def _debug_templates_(self):
        msg = '+++ templates +++'
        for index, template in enumerate(self._templates.get_templates()):
            msg += '\ntemplate #' + str(index) + '\n'
            msg += json.dumps(template, indent=4)
        msg += '\n'
        return msg

    def _debug_spots_(self):
        msg = '+++ spots +++'
        for index, spot in enumerate(self._spots):
            msg += '\nspot #' + str(index) + '\n'
            msg += json.dumps(spot, indent=4)
        msg += '\n'
        return msg

    def _debug_jsonvars_(self):
        msg = '+++ jsonvars +++\n'
        msg += str(self._jsonvars)
        msg += '\n'
        return msg



from ansible import context
from ansible.module_utils.common.collections import ImmutableDict
from ansible.parsing.dataloader import DataLoader
from ansible.playbook.play import Play
from ansible.vars.manager import VariableManager
from takeltest.ansiblerun import AnsibleRun


class MoleculePlay(object):
    '''Run ansible playbooks against molecule host using the ansible python api.
    '''

    def __init__(
            self,
            ansibleinventory,
            moleculeenv):
        # Leverage the ansible python api
        # to run a playbook against a molecule host.
        #
        # see: ansible python api
        # https://docs.ansible.com/ansible/latest/dev_guide/developing_api.html

        self._moleculeenv = moleculeenv

        context.CLIARGS = ImmutableDict(
            connection='local',
            module_path=[''],
            forks=10,
            become=None,
            become_method=None,
            become_user=None,
            check=False,
            diff=False)
        self._loader = DataLoader()
        self._inventory = ansibleinventory
        self._variable_manager = VariableManager(
            loader=DataLoader(),
            inventory=ansibleinventory)

    def get_molecule_scenario_directory(self):
        return self._moleculeenv.get_molecule_scenario_directory()

    def get_roles(self):
        return self._moleculeenv.get_roles()

    def run_playbook(self, playbook):
        '''Run an ansible playbook using the ansible python api.'''
        play = self._get_play_(playbook)
        ansiblerun = AnsibleRun(
            self._inventory,
            self._variable_manager,
            self._loader)
        result = ansiblerun.run(play)
        return result

    def _get_play_(self, playbook):
        return Play().load(
            playbook,
            variable_manager=self._variable_manager,
            loader=self._loader)



import os
import testinfra.utils.ansible_runner


def hosts():

    # if pytest is invoked directly (e.g. to run pytest --cache-clear)
    # molecule is not available so testinfra_hosts should be empty
    if 'MOLECULE_INVENTORY_FILE' in os.environ:
        return testinfra.utils.ansible_runner.AnsibleRunner(
            os.environ['MOLECULE_INVENTORY_FILE']).get_hosts('all')
    else:
        return []
